from .columns_configuration import get_columns_configuration
from .advance_payments_columns import get_advance_payments_columns
from .last_payment_columns import get_last_payment_columns
from .no_payment_due_columns import get_no_payment_due_columns
from .bank_debits_columns import get_bank_debit_columns


from .customer_row_extractor import customer_row_extractor

from .collection_row_extractor import collection_row_extractor
from .collections_extractor_results import CollectionsExtractorResults

from .bill_row_extractor import bill_row_extractor
from .bill_extractor_results import BillExtractorResults

from .account_receivable_extractor import account_receivable_extractor
from .account_receivable_extractor_results import AccountReceivableExtractorResults


from datetime import datetime, timedelta


class CalendarOperations:
    def add_months(self, since_date, months_to_add):

        if not isinstance(since_date, datetime):
            raise ValueError("since_date is not a datetime object")

        if not isinstance(months_to_add, int):
            raise ValueError("months_to_add is not an integer")

        day = since_date.day
        month = since_date.month
        year = since_date.year
        years_to_add = 0
        add_months = months_to_add

        if months_to_add > 12:
            years_to_add = months_to_add // 12
            add_months = months_to_add % 12

        total_months = month + add_months

        if total_months > 12:
            years_to_add = years_to_add + 1
            total_months = total_months - 12

        if total_months == 2 and day > 28:
            day = 28

        if total_months in [4, 6, 9, 11] and day > 30:
            day = 30

        return datetime(year + years_to_add, total_months, day)

    def last_day_of_month(self, date_of_month=datetime.now()):
        next_month = self.add_months(date_of_month, 1)
        first_day_of_next_month = datetime(next_month.year, next_month.month, 1)
        return first_day_of_next_month + timedelta(days=-1)

    def first_day_of_month(self, date_of_month=datetime.now()):
        return datetime(date_of_month.year, date_of_month.month, 1)

    def list_of_due_date(self, since_date, plan):
        return [self.add_months(since_date, x) for x in range(plan)]




import time
import os

from .errornotification import ErrorNotification

from .filenamebuilder import build_file_name

from .readdebitsexcel import read_debits_excel

from .parsers import (
    parse_debits_headers, parse_debits_lines,
    calculate_last_line
)

from .validators import (
    validate_headers, validate_lines,
    validate_file
)

from .notifyerrors import notify_errors

from .writers import (
    stream_lines, stream_last_line
)


def generate_debits(cwd):
    for notification in execute_generation(cwd):
        if notification.has_errors():
            handle_errors(cwd, notification)
            return


def handle_errors(cwd, notification):
    errors_filename = os.path.join(cwd,'outputs/debit_errors_' + time.strftime("%Y%m%d-%H%M%S") + '.txt')
    with open(errors_filename, 'w') as stream:
        notify_errors(notification, stream)


def execute_generation(cwd):
    notification = ErrorNotification()

    excel_path = os.path.join(cwd, "inputs", "generar_debitos.xlsx")

    notification.collect_notifications(validate_file(excel_path))

    yield notification

    raw_data = read_debits_excel(excel_path)

    # procesar cabeceras de excel
    for page_data in raw_data:

        header_data = parse_debits_headers(page_data)
        notification.collect_notifications(validate_headers(header_data))
        yield notification

        debits_data = parse_debits_lines(page_data)
        notification.collect_notifications(validate_lines(debits_data))
        yield notification

        last_line = calculate_last_line(header_data, debits_data)

        filename = build_file_name(cwd, header_data)

        with open(filename, 'w') as file_stream:
            stream_lines(header_data, debits_data, file_stream)
            stream_last_line(last_line, file_stream)



from datetime import datetime
import os

def build_file_name(cwd, header_data):
    filename = os.path.join(cwd, 'outputs', 'ORI' + datetime.utcnow().strftime("%Y%m%d%H%M%S%f")[:-3] + '.txt')

    return filename


import argparse
import os
import sys
import time
from datetime import datetime
import pkg_resources

from functools import reduce

import credisur.dateintelligence as dateintelligence
import credisur.exceladapter as exceladapter
import credisur.excelbuilder as excelbuilder
import credisur.tableextraction as tableextraction

from credisur.config import (
    get_columns_configuration,
    get_advance_payments_columns,
    get_last_payment_columns,
    get_no_payment_due_columns,
    get_bank_debit_columns)


from credisur.extractors import (
    customer_row_extractor,
    collection_row_extractor, CollectionsExtractorResults,
    bill_row_extractor, BillExtractorResults,
    account_receivable_extractor, AccountReceivableExtractorResults
)

from credisur.debitsgenerator import generate_debits
from credisur.bankparser import parse_bank_files

# TODO: Copiar listado de facturas en solapa.
# TODO: Controlar consistencia código (por ejemplo, D-E está mal.
# TODO: Resolver punitorio


def working_directory():
    return os.getcwd().replace('\\', '/')

def valid_date(s):
    try:
        return datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        msg = "No es una fecha válida: '{0}'.".format(s)
        raise argparse.ArgumentTypeError(msg)

def get_parser():
    parser = argparse.ArgumentParser(description='Script para procesar planillas de Credisur')

    parser.add_argument("--version", "-v", help="Muestra la versión.", action="store_true")

    parser.add_argument("--banco", "-b", help="Procesa los archivos del banco disponibles en la carpeta.",
                        action="store_true")

    parser.add_argument("--debitos", "-D", help="Genera los archivos TXT de débitos para enviar al banco.",
                        action="store_true")

    parser.add_argument("--inputs", "-i", help="Permite definir la carpeta de inputs", default="inputs")
    parser.add_argument("--outputs", "-o", help="Permite definir la carpeta de outputs", default="outputs")
    parser.add_argument(
        "--date","-d",
        help="Permite definir la fecha de cálculo para vencimientos. Formato: AAAA-MM-DD. Si no se especifica, toma la fecha de hoy.",
        type=valid_date
    )

    return parser


def is_code_permitted(code):
    return True


def main(args=None):
    PAYMENT_ERRORS = []

    """The main routine."""
    if args is None:
        args = sys.argv[1:]

    cwd = working_directory()

    parser = get_parser()
    params = parser.parse_args(args)

    if params.version:
        print(pkg_resources.require("credisuretl")[0].version)
        return

    if params.banco:
        parse_bank_files(cwd)
        return

    if params.debitos:
        generate_debits(cwd)
        return

    inputs_path = "%s/%s/" % (cwd, params.inputs)
    outputs_path = "%s/%s/" % (cwd, params.outputs)

    calendar_ops = dateintelligence.CalendarOperations()

    # calendar
    current_date = params.date or datetime.now()
    first_day_of_current_month = datetime(current_date.year, current_date.month, 1)
    last_date_of_month = calendar_ops.last_day_of_month(current_date)

    def get_old_excel_filename(filename):
        return filename[:-1]

    def upgrade_if_older_version(filename):
        old_filename = get_old_excel_filename(filename)
        if not os.path.isfile(filename) and os.path.isfile(old_filename):
            exceladapter.ExcelUpgrader(old_filename).upgrade()

    errors = []

    customers_in_last_payment = []
    customers_without_payments_due = []
    list_of_advance_payments = []

    input_customers_filename = inputs_path + 'Clientes.xlsx'
    input_collections_filename = inputs_path + 'Cobranza.xlsx'
    input_pending_bills_filename = inputs_path + 'Factura.xlsx'
    input_accounts_to_collect_filename = inputs_path + 'Cuentas a Cobrar.xlsx'

    upgrade_if_older_version(input_customers_filename)
    upgrade_if_older_version(input_collections_filename)
    upgrade_if_older_version(input_pending_bills_filename)
    upgrade_if_older_version(input_accounts_to_collect_filename)

    # ExcelReader debería tomar un Stream en vez de un filename - TODO: probar

    customer_extractor = tableextraction.DataExtractor(
        input_customers_filename,
        'hoja1',
        {},
        customer_row_extractor
    )

    customers = customer_extractor.extract()

    collection_extractor = tableextraction.DataExtractor(
        input_collections_filename,
        'hoja1',
        CollectionsExtractorResults(),
        collection_row_extractor
    )

    collection_extraction_results = collection_extractor.extract()

    collections = collection_extraction_results.get_collections()
    collections_for_customers = collection_extraction_results.get_collections_for_customers()
    errors = errors + collection_extraction_results.get_errors()

    for customername, customerdetails in customers.items():
        if not customername in collections_for_customers:
            customers_in_last_payment.append({
                "city": customerdetails['city'],
                "customer": customername,
                "address": customerdetails['address'] or 'Sin dirección',
                "lastcollection": "No disponible",
                "reason": "Sin compras abiertas"
            })

    bill_extractor = tableextraction.DataExtractor(
        input_pending_bills_filename,
        'hoja1',
        BillExtractorResults(),
        bill_row_extractor
    )

    bill_extractor_results = bill_extractor.extract()

    bills = bill_extractor_results.get_bills()
    errors = errors + bill_extractor_results.get_errors()

    acc_rec_extractor = tableextraction.DataExtractor(
        input_accounts_to_collect_filename,
        'hoja1',
        AccountReceivableExtractorResults(
            customers, bills,
            calendar_ops, first_day_of_current_month, last_date_of_month,
            collections, collections_for_customers,
            list_of_advance_payments, customers_without_payments_due, customers_in_last_payment
        ),
        account_receivable_extractor
    )

    account_receivable_results = acc_rec_extractor.extract()
    accounts_to_collect = account_receivable_results.get_accounts_to_collect()
    errors = errors + account_receivable_results.get_errors()


    all_customers_with_collection = set(
        list(
            map(
                lambda line: line['customer'],
                reduce(
                    lambda memo, item: memo + item,
                    map(
                        lambda x: x[1],
                        accounts_to_collect.items()
                    )
                )
            )
        )
    )

    for customername, customerdetails in customers.items():
        if not customername in collections_for_customers:
            continue

        if not customername in all_customers_with_collection:
            customers_in_last_payment.append({
                "city": customerdetails['city'],
                "customer": customername,
                "address": customerdetails['address'] or 'Sin dirección',
                "lastcollection": "No disponible",
                "reason": "Sin compras abiertas"
            })

    for error in errors:
        print("ADVERTENCIA:", error)

    sorted_accounts_C = list(sorted(accounts_to_collect['C'],
                               key=lambda x: (x['city'], x['customer'], x['order'], x['due_date_datetime'])))

    sorted_accounts_D = list(sorted(accounts_to_collect['D'],
                               key=lambda x: (x['city'], x['customer'], x['order'], x['due_date_datetime'])))
    sorted_accounts_D_H = list(filter(lambda x: x['person'] == 'H', sorted_accounts_D))
    sorted_accounts_D_F = list(filter(lambda x: x['person'] == 'F', sorted_accounts_D))
    sorted_accounts_I = list(sorted(accounts_to_collect['I'],
                               key=lambda x: (x['city'], x['customer'], x['order'], x['due_date_datetime'])))



    def aggregate_amounts_to_collect(receivables):
        result = {}

        for receivable in receivables:
            customer = receivable['customer']
            amount = receivable['amount_to_collect']
            payment = receivable['payment']
            last_collection = receivable['last_collection']

            if not customer in result:
                cbu = customers[customer]['cbu']
                if not cbu:
                    cbu = ""

                cbu1 = cbu[0:8]
                cbu2 = cbu[8:]
                ref_credit = None

                result[customer] = {}
                result[customer]['event_type'] = 'D'
                result[customer]['customer'] = customer
                result[customer]['cbu1'] = cbu1
                result[customer]['cbu2'] = cbu2
                result[customer]['ref_credit'] = ref_credit
                result[customer]['amount'] = 0
                result[customer]['total_payment'] = 0
                result[customer]['last_collection'] = last_collection

            result[customer]['total_payment'] += amount
            result[customer]['amount'] += min(amount, payment)

        result = list(result.values())

        return result

    receivables_aggregated_by_customer_D_H = aggregate_amounts_to_collect(sorted_accounts_D_H)
    receivables_aggregated_by_customer_D_F = aggregate_amounts_to_collect(sorted_accounts_D_F)
    receivables_aggregated_by_customer_I = aggregate_amounts_to_collect(sorted_accounts_I)
    
    # crear excel de cobranzas
    collections_filename = outputs_path + 'cuentas_a_cobrar_' + time.strftime("%Y%m%d-%H%M%S") + '.xlsx'
    collections_excelwriter = exceladapter.ExcelWriter(collections_filename)

    columns_config = get_columns_configuration()

    collections_builder_C = excelbuilder.BasicBuilder(sorted_accounts_C, columns_config)
    collections_excelwriter.build_sheet("Créditos", collections_builder_C.build_sheet_data())

    collections_builder_DH = excelbuilder.BasicBuilder(sorted_accounts_D_H, columns_config)
    collections_excelwriter.build_sheet('Débitos Horacio', collections_builder_DH.build_sheet_data())

    collections_builder_DF = excelbuilder.BasicBuilder(sorted_accounts_D_F, columns_config)
    collections_excelwriter.build_sheet('Débitos Facundo', collections_builder_DF.build_sheet_data())

    collections_builder_I = excelbuilder.BasicBuilder(sorted_accounts_I, columns_config)
    collections_excelwriter.build_sheet('ICBC', collections_builder_I.build_sheet_data())

    bank_columns_config = get_bank_debit_columns()

    bank_debit_builder_DH = excelbuilder.BasicBuilder(receivables_aggregated_by_customer_D_H, bank_columns_config)

    bank_debit_builder_DH.set_header_row(7)
    bank_debit_builder_DH.set_first_data_row(8)

    bank_debit_builder_DH.set_manual_cell("A1", "CUIT")
    bank_debit_builder_DH.set_manual_cell("B1", "27316719517")

    # Horacio: CREDISURII
    bank_debit_builder_DH.set_manual_cell("A2", "Prestación")
    bank_debit_builder_DH.set_manual_cell("B2", "CREDISURII")

    # CREDISUR
    bank_debit_builder_DH.set_manual_cell("A3", "Ref del Débito")
    bank_debit_builder_DH.set_manual_cell("B3", "CREDISUR")

    bank_debit_builder_DH.set_manual_cell("A4", "Fecha 1er vencimiento")
    bank_debit_builder_DH.set_manual_cell("A5", "Fecha de proceso")
    bank_debit_builder_DH.add_header("F", "Observaciones")

    collections_excelwriter.build_sheet('Banco Horacio', bank_debit_builder_DH.build_sheet_data())

    bank_debit_builder_DF = excelbuilder.BasicBuilder(receivables_aggregated_by_customer_D_F, bank_columns_config)
    bank_debit_builder_DF.set_header_row(7)
    bank_debit_builder_DF.set_first_data_row(8)

    bank_debit_builder_DF.set_manual_cell("A1", "CUIT")
    bank_debit_builder_DF.set_manual_cell("B1", "20292879742")

    # Facundo: V.MUEBLES
    bank_debit_builder_DF.set_manual_cell("A2", "Prestación")
    bank_debit_builder_DF.set_manual_cell("B2", "V.MUEBLES")

    # CREDISUR
    bank_debit_builder_DF.set_manual_cell("A3", "Ref del Débito")
    bank_debit_builder_DF.set_manual_cell("B3", "CREDISUR")

    bank_debit_builder_DF.set_manual_cell("A4", "Fecha 1er vencimiento")
    bank_debit_builder_DF.set_manual_cell("A5", "Fecha de proceso")
    bank_debit_builder_DF.add_header("F", "Observaciones")

    collections_excelwriter.build_sheet('Banco Facundo', bank_debit_builder_DF.build_sheet_data())

    bank_debit_builder_I = excelbuilder.BasicBuilder(receivables_aggregated_by_customer_I, bank_columns_config)
    bank_debit_builder_I.set_header_row(7)
    bank_debit_builder_I.set_first_data_row(8)

    bank_debit_builder_I.set_manual_cell("A1", "CUIT")
    bank_debit_builder_I.set_manual_cell("B1", "20292879742")

    # ICBC: V.MUEBLES
    bank_debit_builder_I.set_manual_cell("A2", "Prestación")
    bank_debit_builder_I.set_manual_cell("B2", "V.MUEBLES")

    # CREDISUR
    bank_debit_builder_I.set_manual_cell("A3", "Ref del Débito")
    bank_debit_builder_I.set_manual_cell("B3", "D MUEBLES")

    bank_debit_builder_I.set_manual_cell("A4", "Fecha 1er vencimiento")
    bank_debit_builder_I.set_manual_cell("A5", "Fecha de proceso")
    bank_debit_builder_I.add_header("F", "Observaciones")

    collections_excelwriter.build_sheet('Banco ICBC', bank_debit_builder_I.build_sheet_data())

    collections_excelwriter.save()

    last_payment_filename = outputs_path + 'última_cuota_' + time.strftime("%Y%m%d-%H%M%S") + '.xlsx'

    last_payment_composer = exceladapter.ExcelBasicComposer(
        last_payment_filename,
        customers_in_last_payment,
        get_last_payment_columns(),
        "Última Cuota"
    )
    last_payment_composer.save()


    no_payment_due_filename = outputs_path + 'ordenes_de_compra_abiertas_sin_cuotas_a_cobrar_este_periodo' + time.strftime(
        "%Y%m%d-%H%M%S") + '.xlsx'

    no_payment_due_composer = exceladapter.ExcelBasicComposer(
        no_payment_due_filename,
        customers_without_payments_due,
        get_no_payment_due_columns(),
        "Sin cuotas"
    )
    no_payment_due_composer.save()

    advance_payments_filename = outputs_path + 'anticipos_' + time.strftime("%Y%m%d-%H%M%S") + '.xlsx'

    advance_payments_composer = exceladapter.ExcelBasicComposer(
        advance_payments_filename,
        list_of_advance_payments,
        get_advance_payments_columns(),
        "Anticipos"
    )
    advance_payments_composer.save()

    errors_filename = outputs_path + 'errors_' + time.strftime("%Y%m%d-%H%M%S") + '.txt'
    with open(errors_filename, 'w') as f:
        for error in errors:
            f.write(error)
            f.write("\n")

    for payment_error in PAYMENT_ERRORS:
        print(payment_error)


if __name__ == "__main__":
    main()



from .errorsiterator import ErrorsIterator

def validate_lines(debits_data):
    errors = ErrorsIterator()

    validate_func = validate_line_func(errors)

    for line in debits_data:
        validate_func(line)

    return errors

def validate_line_func(errors):

    fields = {
        'tipo_novedad': 'Tipo Novedad',
        'cbu_bloque_1': 'CBU Bloque 1',
        'cbu_bloque_2': 'CBU Bloque 2',
        'id_cliente': 'ID Cliente',
        'monto': 'Monto',
    }

    def validate_line(line):
        """

        :type line: credisur.debitsgenerator.parsers.parsedebitslines.DebitLine
        """

        row = line.row
        sheet_name = line.sheet_name

        def add_error(error):
            errors.add_error(error + " en fila %s de solapa %s" % (row, sheet_name))

        check_for_presence(line, fields, add_error)

    return validate_line


def check_for_presence(line, fields, err_func):
    for field, error_message in fields.items():
        if not getattr(line, field):
            err_func("Falta el tipo '%s'" % (error_message,))




import xlrd
from openpyxl.workbook import Workbook
import datetime

class ExcelUpgrader:
    def __init__(self, filename):
        self.filename = filename

    def upgrade(self):
        print(self.filename)

        # first open using xlrd
        book = xlrd.open_workbook(self.filename, formatting_info=True)
        index = 0
        nrows, ncols = 0, 0

        while nrows * ncols == 0:
            sheet = book.sheet_by_index(index)
            nrows = sheet.nrows
            ncols = sheet.ncols
            index += 1

        # prepare a xlsx sheet
        book1 = Workbook()
        sheet1 = book1.get_active_sheet()
        sheet1.title = 'hoja1'

        for row in range(1, nrows+1):
            for col in range(1, ncols+1):
                sheet1.cell(row=row, column=col).value = self._get_xlrd_cell_value(sheet.cell(row-1, col-1))

        book1.save(self.filename + "x")

    def _get_xlrd_cell_value(self,cell):
        value = cell.value
        if cell.ctype == xlrd.XL_CELL_DATE:
            value = datetime.datetime(*xlrd.xldate_as_tuple(value, 0))

        return value


from credisur.models import Collections
from credisur.datastructures import HashOfLists

class CollectionsExtractorResults:

    def __init__(self):
        self.collections = Collections()
        self.collections_for_customers = HashOfLists()
        self.errors = []


    def append_to_collections(self, collection):
        self.collections.append(collection)

    def append_to_collections_for_customers(self, customer, collection):
        self.collections_for_customers.append(customer, collection.to_dict())

    def get_errors(self):
        return self.errors

    def get_collections(self):
        return self.collections

    def get_collections_for_customers(self):
        return self.collections_for_customers
                



def stream_lines(header_data, debits_data, stream):
    for line in parse_lines(header_data, debits_data):
        stream.write(line + "\n")


def parse_lines(header_data, debits_data):
    func_to_map = build_map_func(header_data)
    return map(func_to_map, debits_data)


def build_map_func(header_data):
    """

    :type header_data: credisur.debitsgenerator.parsers.parsedebitsheaders.Header
    """

    def build_string_for(line):
        """

        :type line: credisur.debitsgenerator.parsers.parsedebitslines.DebitLine
        """
        return "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s" % (
            line.tipo_novedad,                      #  1      1      1
            header_data.cuit,                       # 11      2     12
            " " * 3, # sector                       #  3     13     15
            header_data.prestacion.ljust(10),       # 10     16     25
            header_data.primer_vencimiento,         #  8     26     33
            line.cbu_bloque_1,                      #  8     34     41
            "0" * 3, # filler                       #  3     42     44
            line.cbu_bloque_2,                      # 14     45     58
            line.id_cliente[:22].ljust(22),         # 22     59     80
            header_data.primer_vencimiento,         #  8     81     88
            header_data.ref_debito.ljust(15),       # 15     89    103
            line.monto,                             # 10    104    113
            "80", # moneda 80 pesos 82 dólares      #  2    114    115
            " " * 8, # fecha segundo vencimiento    #  8    116    123
            "0" * 10, # monto segundo vencimiento   # 10    124    133
            " " * 8,  # fecha tercer vencimiento    #  8    134    141
            "0" * 10,  # monto tercer vencimiento   # 10    142    151
            " " * 22, # no corresponde              # 22    152    173
            " " * 3, # no corresponde               #  3    174    176
            "0" * 10, # no corresponde              # 10    177    186
            "0" * 10, # no corresponde              # 10    187    196
            " " * 54 # FILLER                       # 54    197    250
        )

    return build_string_for

def stream_last_line(last_line, stream):
    line = parse_last_line(last_line)

    if not line:
        return

    stream.write(line)


def parse_last_line(last_line):
    """

    :type last_line: credisur.debitsgenerator.parsers.lastlinecalculator.LastLine
    """

    return "%s%s%s%s%s%s%s%s" % (
        last_line.tipo_de_novedad,          #  1    1    1
        last_line.registros_totales,        # 10    2   11
        last_line.registros_monetarios,     #  7   12   18
        last_line.registros_no_monetarios,  #  7   19   25
        last_line.fecha_proceso,            #  8   26   33
        " " * 70,                           # 70   34  103
        last_line.suma_importes_registros,  # 70   34  103
        " " * 137                           #137  114  259
    )


from credisur.models import Collection

def collection_row_extractor(results, unpacker):
    document = unpacker.get_value_at(2)
    raw_code = unpacker.get_value_at(5) or ""
    date = unpacker.get_value_at(1)
    customer = unpacker.get_value_at(3)
    amount = unpacker.get_value_at(4)

    collection = Collection(document, raw_code,
                            customer,
                            date, amount,
                            results.get_errors())

    results.append_to_collections(collection)
    results.append_to_collections_for_customers(customer, collection)